dnl  mpn_mul_1 -- Multiply a limb vector with a limb and store
dnl  the result in a second limb vector.

dnl  Copyright 2001, 2002, 2003, 2005 Free Software Foundation, Inc.
dnl
dnl  This file is part of the GNU MP Library.
dnl
dnl  The GNU MP Library is free software; you can redistribute it and/or
dnl  modify it under the terms of the GNU Lesser General Public License as
dnl  published by the Free Software Foundation; either version 3 of the
dnl  License, or (at your option) any later version.
dnl
dnl  The GNU MP Library is distributed in the hope that it will be useful,
dnl  but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
dnl  Lesser General Public License for more details.
dnl
dnl  You should have received a copy of the GNU Lesser General Public License
dnl  along with the GNU MP Library.  If not, see http://www.gnu.org/licenses/.

dnl
dnl *********************************************************************
dnl
dnl  Modified by ct35z, 2008/03/25
dnl

include(`../config.m4')


C mp_limb_t mpn_mul_1 (mp_ptr dst, mp_srcptr src, mp_size_t size,
C                      mp_limb_t multiplier);
C mp_limb_t mpn_mul_1c (mp_ptr dst, mp_srcptr src, mp_size_t size,
C                       mp_limb_t multiplier, mp_limb_t carry);


defframe(PARAM_CARRY,     20)
defframe(PARAM_MULTIPLIER,16)
defframe(PARAM_SIZE,      12)
defframe(PARAM_SRC,       8)
defframe(PARAM_DST,       4)

	TEXT
	ALIGN(16)
PROLOGUE(mpn_mul_1c)
deflit(`FRAME',0)
	movd	PARAM_CARRY, %xmm0
	jmp	L(start_1c)
EPILOGUE()

PROLOGUE(mpn_mul_1)
deflit(`FRAME',0)
	pxor	%xmm0, %xmm0
L(start_1c):
	pushl	%edi
	pushl	%esi
deflit(`FRAME',8)

	movl	PARAM_SIZE, %edx
	movl	PARAM_SRC, %esi
	movd	PARAM_MULTIPLIER, %xmm7
	movl	PARAM_DST, %edi

	movl	%edx,%ecx
	andl	$15,%edx
	movl	%edx,%eax

	movl	L(mul0_jmptbl)(,%edx,4),%edx
	jmp	*%edx

	ALIGN(4)
L(mul0_jmptbl):
	.long	L(mul0_s0)
	.long	L(mul0_s1)
	.long	L(mul0_s2)
	.long	L(mul0_s3)
	.long	L(mul0_s4)
	.long	L(mul0_s5)
	.long	L(mul0_s6)
	.long	L(mul0_s7)
	.long	L(mul0_s8)
	.long	L(mul0_s9)
	.long	L(mul0_s10)
	.long	L(mul0_s11)
	.long	L(mul0_s12)
	.long	L(mul0_s13)
	.long	L(mul0_s14)
	.long	L(mul0_s15)

L(mul0_loop):
	addl	$16,%eax
							C Upper 64bits of xmm registers are not used.
	movd	-64+4*0(%esi,%eax,4), %xmm1
	pmuludq	%xmm7, %xmm1
	paddq	%xmm1, %xmm0
	movd	%xmm0, -64+4*0(%edi,%eax,4)
	psrlq	$32, %xmm0
L(mul0_s15):
	movd	-64+4*1(%esi,%eax,4), %xmm1
	pmuludq	%xmm7, %xmm1
	paddq	%xmm1, %xmm0
	movd	%xmm0, -64+4*1(%edi,%eax,4)
	psrlq	$32, %xmm0
L(mul0_s14):
	movd	-64+4*2(%esi,%eax,4), %xmm1
	pmuludq	%xmm7, %xmm1
	paddq	%xmm1, %xmm0
	movd	%xmm0, -64+4*2(%edi,%eax,4)
	psrlq	$32, %xmm0
L(mul0_s13):
	movd	-64+4*3(%esi,%eax,4), %xmm1
	pmuludq	%xmm7, %xmm1
	paddq	%xmm1, %xmm0
	movd	%xmm0, -64+4*3(%edi,%eax,4)
	psrlq	$32, %xmm0
L(mul0_s12):
	movd	-64+4*4(%esi,%eax,4), %xmm1
	pmuludq	%xmm7, %xmm1
	paddq	%xmm1, %xmm0
	movd	%xmm0, -64+4*4(%edi,%eax,4)
	psrlq	$32, %xmm0
L(mul0_s11):
	movd	-64+4*5(%esi,%eax,4), %xmm1
	pmuludq	%xmm7, %xmm1
	paddq	%xmm1, %xmm0
	movd	%xmm0, -64+4*5(%edi,%eax,4)
	psrlq	$32, %xmm0
L(mul0_s10):
	movd	-64+4*6(%esi,%eax,4), %xmm1
	pmuludq	%xmm7, %xmm1
	paddq	%xmm1, %xmm0
	movd	%xmm0, -64+4*6(%edi,%eax,4)
	psrlq	$32, %xmm0
L(mul0_s9):
	movd	-64+4*7(%esi,%eax,4), %xmm1
	pmuludq	%xmm7, %xmm1
	paddq	%xmm1, %xmm0
	movd	%xmm0, -64+4*7(%edi,%eax,4)
	psrlq	$32, %xmm0
L(mul0_s8):
	movd	-64+4*8(%esi,%eax,4), %xmm1
	pmuludq	%xmm7, %xmm1
	paddq	%xmm1, %xmm0
	movd	%xmm0, -64+4*8(%edi,%eax,4)
	psrlq	$32, %xmm0
L(mul0_s7):
	movd	-64+4*9(%esi,%eax,4), %xmm1
	pmuludq	%xmm7, %xmm1
	paddq	%xmm1, %xmm0
	movd	%xmm0, -64+4*9(%edi,%eax,4)
	psrlq	$32, %xmm0
L(mul0_s6):
	movd	-64+4*10(%esi,%eax,4), %xmm1
	pmuludq	%xmm7, %xmm1
	paddq	%xmm1, %xmm0
	movd	%xmm0, -64+4*10(%edi,%eax,4)
	psrlq	$32, %xmm0
L(mul0_s5):
	movd	-64+4*11(%esi,%eax,4), %xmm1
	pmuludq	%xmm7, %xmm1
	paddq	%xmm1, %xmm0
	movd	%xmm0, -64+4*11(%edi,%eax,4)
	psrlq	$32, %xmm0
L(mul0_s4):
	movd	-64+4*12(%esi,%eax,4), %xmm1
	pmuludq	%xmm7, %xmm1
	paddq	%xmm1, %xmm0
	movd	%xmm0, -64+4*12(%edi,%eax,4)
	psrlq	$32, %xmm0
L(mul0_s3):
	movd	-64+4*13(%esi,%eax,4), %xmm1
	pmuludq	%xmm7, %xmm1
	paddq	%xmm1, %xmm0
	movd	%xmm0, -64+4*13(%edi,%eax,4)
	psrlq	$32, %xmm0
L(mul0_s2):
	movd	-64+4*14(%esi,%eax,4), %xmm1
	pmuludq	%xmm7, %xmm1
	paddq	%xmm1, %xmm0
	movd	%xmm0, -64+4*14(%edi,%eax,4)
	psrlq	$32, %xmm0
L(mul0_s1):
	movd	-64+4*15(%esi,%eax,4), %xmm1
	pmuludq	%xmm7, %xmm1
	paddq	%xmm1, %xmm0
	movd	%xmm0, -64+4*15(%edi,%eax,4)
	psrlq	$32, %xmm0
L(mul0_s0):
L(mul0_start):
	cmpl	%ecx,%eax
	jnz		L(mul0_loop)
L(mul0_end):

	movd	%xmm0, %eax

	popl	%esi
	popl	%edi
	ret

EPILOGUE()

