dnl *********************************************************************
dnl  Intel64 mpn_mul_1 -- Multiply a limb vector with a limb and
dnl  store the result.
dnl
dnl  Copyright 2006  Jason Worth Martin <jason.worth.martin@gmail.com>
dnl
dnl  This code is distributed under the terms of the Gnu Lesser
dnl  General Public License.
dnl
dnl  This code is part of a patch for the GNU MP Library (GMP).
dnl  However, this code is not maintained by the GMP group, so please
dnl  do not bother them with questions/bugs/etc. related to this code.
dnl  If you find a bug in this code, please contact Jason Worth Martin
dnl  at jason.worth.martin@gmail.com
dnl
dnl  This patch is free software; you can redistribute it and/or modify
dnl  it under the terms of the GNU Lesser General Public License as published
dnl  by the Free Software Foundation; either version 2.1 of the License, or (at
dnl  your option) any later version.
dnl
dnl  This patch is distributed in the hope that it will be useful, but
dnl  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
dnl  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
dnl  License for more details.
dnl
dnl  You should have received a copy of the GNU Lesser General Public License
dnl  along with the GNU MP Library; see the file LICENSE.TXT.  If not, write
dnl  to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
dnl  Boston, MA 02110-1301, USA.
dnl
dnl *********************************************************************
dnl
dnl  Modified by ct35z, 2008/03/22
dnl
dnl *********************************************************************
dnl
dnl
dnl CREDITS
dnl
dnl This code is based largely on Pierrick Gaudry's excellent assembly
dnl support for the AMD64 architecture.  (Note that Intel64 and AMD64,
dnl while using the same instruction set, have very different
dnl microarchitectures.  So, this code performs very poorly on AMD64
dnl machines even though it is near-optimal on Intel64.)
dnl
dnl Roger Golliver works for Intel and provided insightful improvements
dnl particularly in using the "lea" instruction to perform additions
dnl and register-to-register moves.
dnl
dnl Eric Bainville has a brilliant exposition of optimizing arithmetic for
dnl AMD64 (http://www.bealto.it).  I adapted many of the ideas he
dnl describes to Intel64.
dnl
dnl Agner Fog is a demigod in the x86 world.  If you are reading assembly
dnl code files and you haven't heard of Agner Fog, then take a minute to
dnl look over his software optimization manuals (http://www.agner.org/).
dnl They are superb.
dnl


dnl
dnl This is just a check to see if we are in my code testing sandbox
dnl or if we are actually in the GMP source tree
dnl
ifdef(`__JWM_Test_Code__',`
include(`./config.m4')
define(`MPN_PREFIX',`jwm_mpn_')',`
include(`../config.m4')')


dnl *********************************************************************
dnl *********************************************************************
dnl
dnl     BpL is Bytes per Limb (8 since this is 64bit code)
dnl
dnl *********************************************************************
dnl *********************************************************************
define(`BpL',`8')
define(`UNROLL_EXPONENT',`4')
define(`UNROLL_SIZE',eval(2**UNROLL_EXPONENT))
define(`UNROLL_MASK',eval(UNROLL_SIZE-1))

dnl Here is a convenient Macro for addressing
dnl memory.  Entries of the form
dnl
dnl      ADDR(ptr,index,displacement)
dnl
dnl get converted to
dnl
dnl      displacement*BpL(ptr,index,BpL)
dnl
define(`ADDR',`eval(`$3'*BpL)($1,$2,BpL)')



C Register	Usage
C --------	-----
C rax		low word from mul
C rbx*
C rcx		s2limb
C rdx		high word from mul
C rsi		s1p
C rdi		rp
C rbp*		Base Pointer
C rsp*		Stack Pointer
C r8		A_x
C r9		A_y
C r10		A_z
C r11		B_x
C r12*		B_y
C r13*		B_z
C r14*		temp
C r15*		index
C
C * indicates that the register must be
C preserved for the caller.
define(`s2limb',`%rcx')
define(`s1p',`%rsi')
define(`rp',`%rdi')
define(`A_x',`%r8')
define(`A_y',`%r9')
define(`A_z',`%r10')
define(`B_x',`%r11')
define(`B_y',`%r12')
define(`B_z',`%r13')
define(`temp',`%r14')
define(`index',`%r15')

	
C INPUT PARAMETERS
C rp		rdi
C s1p		rsi
C n		rdx
C s2limb	rcx
	
ASM_START()
PROLOGUE(mpn_mul_1)
	pushq	%r15
	pushq	%r14
	pushq	%r13
	pushq	%r12

	leaq	ADDR(s1p,%rdx,0), s1p
	leaq	ADDR(rp,%rdx,0), rp
	xorq	index, index
	subq	%rdx, index

	cmp	$4,%rdx
	jge	L_mpn_mul_1_main_loop_prep


	leaq	L_mpn_mul_1_jmptbl0(%rip), %rax
	addq	(%rax, %rdx, 8), %rax
	jmp	*%rax

	ALIGN(8)
L_mpn_mul_1_jmptbl0:
	.quad	L_mpn_mul_1_small_0 - L_mpn_mul_1_jmptbl0
	.quad	L_mpn_mul_1_small_1 - L_mpn_mul_1_jmptbl0
	.quad	L_mpn_mul_1_small_2 - L_mpn_mul_1_jmptbl0
	.quad	L_mpn_mul_1_small_3 - L_mpn_mul_1_jmptbl0


L_mpn_mul_1_small_0:
	xorq	%rax, %rax
	jmp	L_mpn_mul_1_exit

L_mpn_mul_1_small_1:
	movq	ADDR(s1p,index,0), %rax
	mulq	s2limb
	movq	%rax, ADDR(rp,index,0)
	movq	%rdx, %rax
	jmp	L_mpn_mul_1_exit

L_mpn_mul_1_small_2:
	movq	ADDR(s1p,index,0), %rax
	mulq	s2limb
	movq	%rax, A_x
	movq	%rdx, A_y

	movq	ADDR(s1p,index,1), %rax
	mulq	s2limb

	addq	A_y, %rax
	movq	A_x, ADDR(rp,index,0)
	adcq	$0, %rdx

	movq	%rax, ADDR(rp,index,1)

	movq	%rdx, %rax
	jmp	L_mpn_mul_1_exit

L_mpn_mul_1_small_3:
	movq	ADDR(s1p,index,0), %rax
	mulq	s2limb
	movq	%rax, A_x
	movq	%rdx, A_y

	movq	ADDR(s1p,index,1), %rax
	mulq	s2limb
	movq	%rax, B_x
	movq	%rdx, B_y

	movq	ADDR(s1p,index,2), %rax

	mulq	s2limb
	addq	A_y, B_x
	movq	A_x, ADDR(rp,index,0)
	adcq	$0, B_y

	addq	B_y, %rax
	movq	B_x, ADDR(rp,index,1)
	adcq	$0, %rdx

	movq	%rax, ADDR(rp,index,2)

	movq	%rdx, %rax
	jmp	L_mpn_mul_1_exit




L_mpn_mul_1_main_loop_prep:
	movq	%rdx, temp

	testq	$1, %rdx
	jz	L_mpn_mul_1_main_loop_prep_n1

	movq	ADDR(s1p,index,0), %rax
	mulq	s2limb
	movq	%rax, A_x
	movq	%rdx, A_y

	movq	ADDR(s1p,index,1), %rax
	mulq	s2limb
	movq	%rax, B_x
	movq	%rdx, B_y

	jmp	L_mpn_mul_1_main_loop_prep_n2

L_mpn_mul_1_main_loop_prep_n1:
	movq	ADDR(s1p,index,0), %rax
	mulq	s2limb
	movq	%rax, B_x
	movq	%rdx, B_y

	movq	ADDR(s1p,index,1), %rax
	mulq	s2limb
	movq	%rax, A_x
	movq	%rdx, A_y

L_mpn_mul_1_main_loop_prep_n2:

	movq	ADDR(s1p,index,2), %rax

	subq	$4, temp
	andq	$UNROLL_MASK, temp
	lea	(4-UNROLL_SIZE)(index, temp), index

	leaq	L_mpn_mul_1_jmptbl(%rip), %rdx
	addq	(%rdx, temp, 8), %rdx
	jmp	*%rdx


	ALIGN(8)
L_mpn_mul_1_jmptbl:
	.quad	L_mpn_mul_1_main_loop_s1 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s2 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s3 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s4 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s5 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s6 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s7 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s8 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s9 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s10 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s11 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s12 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s13 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s14 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s15 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s16 - L_mpn_mul_1_jmptbl


	ALIGN(16)
L_mpn_mul_1_main_loop:

L_mpn_mul_1_main_loop_s16:
	mulq	s2limb
	movq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*0)), %rax
	addq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*0-3))
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_mul_1_main_loop_s15:
	mulq	s2limb
	movq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*0+1)), %rax
	addq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*0-2))
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_mul_1_main_loop_s14:
	mulq	s2limb
	movq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*1)), %rax
	addq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*1-3))
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_mul_1_main_loop_s13:
	mulq	s2limb
	movq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*1+1)), %rax
	addq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*1-2))
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_mul_1_main_loop_s12:
	mulq	s2limb
	movq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*2)), %rax
	addq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*2-3))
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_mul_1_main_loop_s11:
	mulq	s2limb
	movq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*2+1)), %rax
	addq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*2-2))
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_mul_1_main_loop_s10:
	mulq	s2limb
	movq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*3)), %rax
	addq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*3-3))
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_mul_1_main_loop_s9:
	mulq	s2limb
	movq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*3+1)), %rax
	addq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*3-2))
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_mul_1_main_loop_s8:
	mulq	s2limb
	movq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*4)), %rax
	addq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*4-3))
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_mul_1_main_loop_s7:
	mulq	s2limb
	movq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*4+1)), %rax
	addq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*4-2))
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_mul_1_main_loop_s6:
	mulq	s2limb
	movq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*5)), %rax
	addq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*5-3))
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_mul_1_main_loop_s5:
	mulq	s2limb
	movq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*5+1)), %rax
	addq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*5-2))
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_mul_1_main_loop_s4:
	mulq	s2limb
	movq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*6)), %rax
	addq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*6-3))
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_mul_1_main_loop_s3:
	mulq	s2limb
	movq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*6+1)), %rax
	addq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*6-2))
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_mul_1_main_loop_s2:
	mulq	s2limb
	movq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*7)), %rax
	addq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*7-3))
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_mul_1_main_loop_s1:
	mulq	s2limb
	movq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*7+1)), %rax
	addq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*7-2))
	leaq	(%rdx), B_y
	adcq	$0, A_y

	addq	$UNROLL_SIZE, index
	jnz	L_mpn_mul_1_main_loop

L_mpn_mul_1_finish:	

	mulq	s2limb
	movq	A_x, ADDR(rp,index,-3)
	addq	A_y, B_x
	adcq	$0, B_y

	movq	B_x, ADDR(rp,index,-2)
	addq	B_y, %rax
	adcq	$0, %rdx

	movq	%rax, ADDR(rp,index,-1)

	movq	%rdx, %rax
L_mpn_mul_1_exit:
	popq	%r12
	popq	%r13
	popq	%r14
	popq	%r15
	ret
EPILOGUE()
