dnl  Intel64 mpn_mul_basecase -- mpn by mpn multiplication.
dnl
dnl  Copyright 2001, 2002 Free Software Foundation, Inc.
dnl
dnl  This file is part of the GNU MP Library.
dnl
dnl  The GNU MP Library is free software; you can redistribute it and/or
dnl  modify it under the terms of the GNU Lesser General Public License as
dnl  published by the Free Software Foundation; either version 2.1 of the
dnl  License, or (at your option) any later version.
dnl
dnl  The GNU MP Library is distributed in the hope that it will be useful,
dnl  but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
dnl  Lesser General Public License for more details.
dnl
dnl  You should have received a copy of the GNU Lesser General Public
dnl  License along with the GNU MP Library; see the file COPYING.LIB.  If
dnl  not, write to the Free Software Foundation, Inc., 51 Franklin Street,
dnl  Fifth Floor, Boston, MA 02110-1301, USA.
dnl
dnl
dnl *********************************************************************
dnl
dnl  Modified by ct35z, 2008/03/22
dnl
dnl *********************************************************************
dnl
dnl
dnl CREDITS
dnl
dnl This code is based largely on Jason Worth Martin's excellent assembly
dnl support for the Intel64 architecture.
dnl


dnl
dnl This is just a check to see if we are in my code testing sandbox
dnl or if we are actually in the GMP source tree
dnl
ifdef(`__JWM_Test_Code__',`
include(`./config.m4')
define(`MPN_PREFIX',`jwm_mpn_')',`
include(`../config.m4')')


dnl *********************************************************************
dnl *********************************************************************
dnl
dnl     BpL is Bytes per Limb (8 since this is 64bit code)
dnl
dnl *********************************************************************
dnl *********************************************************************
define(`BpL',`8')
define(`UNROLL_EXPONENT',`4')
define(`UNROLL_SIZE',eval(2**UNROLL_EXPONENT))
define(`UNROLL_MASK',eval(UNROLL_SIZE-1))

dnl Here is a convenient Macro for addressing
dnl memory.  Entries of the form
dnl
dnl      ADDR(ptr,index,displacement)
dnl
dnl get converted to
dnl
dnl      displacement*BpL(ptr,index,BpL)
dnl
define(`ADDR',`eval(`$3'*BpL)($1,$2,BpL)')



define(`s2limb',`%rbp')
define(`s1p',`%rsi')
define(`rp',`%rdi')
define(`A_x',`%rbx')
define(`A_y',`%r9')
define(`A_z',`%r10')
define(`B_x',`%r11')
define(`B_y',`%r12')
define(`B_z',`%r13')
define(`un',`%r14')
define(`index',`%r15')
define(`vn',`%r8')
define(`vp',`%rcx')

	

C INPUT PARAMETERS
C rp	rdi	PARAM_WP
C up	rsi	PARAM_XP
C un	rdx	PARAM_XSIZE
C vp	rcx	PARAM_YP
C vn	r8	PARAM_YSIZE


ASM_START()
PROLOGUE(mpn_mul_basecase)
	pushq	%r15
	pushq	%r14
	pushq	%r13
	pushq	%r12
	pushq	%rbx
	pushq	%rbp

	movq	%rdx, un
	movq	(vp), s2limb


L_mpn_mul_1:
	leaq	ADDR(s1p,%rdx,0), s1p
	leaq	ADDR(rp,%rdx,0), rp
	xorq	index, index
	subq	%rdx, index

	cmp	$4,%rdx
	jge	L_mpn_mul_1_main_loop_prep


	leaq	L_mpn_mul_1_jmptbl0(%rip), %rax
	addq	(%rax, %rdx, 8), %rax
	jmp	*%rax

	ALIGN(8)
L_mpn_mul_1_jmptbl0:
	.quad	L_mpn_mul_1_small_0 - L_mpn_mul_1_jmptbl0
	.quad	L_mpn_mul_1_small_1 - L_mpn_mul_1_jmptbl0
	.quad	L_mpn_mul_1_small_2 - L_mpn_mul_1_jmptbl0
	.quad	L_mpn_mul_1_small_3 - L_mpn_mul_1_jmptbl0


L_mpn_mul_1_small_0:
	xorq	%rax, %rax
	jmp	L_mpn_mul_1_exit

L_mpn_mul_1_small_1:
	movq	ADDR(s1p,index,0), %rax
	mulq	s2limb
	movq	%rax, ADDR(rp,index,0)
	movq	%rdx, %rax
	jmp	L_mpn_mul_1_exit

L_mpn_mul_1_small_2:
	movq	ADDR(s1p,index,0), %rax
	mulq	s2limb
	movq	%rax, A_x
	movq	%rdx, A_y

	movq	ADDR(s1p,index,1), %rax
	mulq	s2limb

	addq	A_y, %rax
	movq	A_x, ADDR(rp,index,0)
	adcq	$0, %rdx

	movq	%rax, ADDR(rp,index,1)

	movq	%rdx, %rax
	jmp	L_mpn_mul_1_exit

L_mpn_mul_1_small_3:
	movq	ADDR(s1p,index,0), %rax
	mulq	s2limb
	movq	%rax, A_x
	movq	%rdx, A_y

	movq	ADDR(s1p,index,1), %rax
	mulq	s2limb
	movq	%rax, B_x
	movq	%rdx, B_y

	movq	ADDR(s1p,index,2), %rax

	mulq	s2limb
	addq	A_y, B_x
	movq	A_x, ADDR(rp,index,0)
	adcq	$0, B_y

	addq	B_y, %rax
	movq	B_x, ADDR(rp,index,1)
	adcq	$0, %rdx

	movq	%rax, ADDR(rp,index,2)

	movq	%rdx, %rax
	jmp	L_mpn_mul_1_exit




L_mpn_mul_1_main_loop_prep:
	movq	%rdx, A_z

	testq	$1, %rdx
	jz	L_mpn_mul_1_main_loop_prep_n1

	movq	ADDR(s1p,index,0), %rax
	mulq	s2limb
	movq	%rax, A_x
	movq	%rdx, A_y

	movq	ADDR(s1p,index,1), %rax
	mulq	s2limb
	movq	%rax, B_x
	movq	%rdx, B_y

	jmp	L_mpn_mul_1_main_loop_prep_n2

L_mpn_mul_1_main_loop_prep_n1:
	movq	ADDR(s1p,index,0), %rax
	mulq	s2limb
	movq	%rax, B_x
	movq	%rdx, B_y

	movq	ADDR(s1p,index,1), %rax
	mulq	s2limb
	movq	%rax, A_x
	movq	%rdx, A_y

L_mpn_mul_1_main_loop_prep_n2:

	movq	ADDR(s1p,index,2), %rax

	subq	$4, A_z
	andq	$UNROLL_MASK, A_z
	lea	(4-UNROLL_SIZE)(index, A_z), index

	leaq	L_mpn_mul_1_jmptbl(%rip), %rdx
	addq	(%rdx, A_z, 8), %rdx
	jmp	*%rdx


	ALIGN(8)
L_mpn_mul_1_jmptbl:
	.quad	L_mpn_mul_1_main_loop_s1 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s2 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s3 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s4 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s5 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s6 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s7 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s8 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s9 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s10 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s11 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s12 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s13 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s14 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s15 - L_mpn_mul_1_jmptbl
	.quad	L_mpn_mul_1_main_loop_s16 - L_mpn_mul_1_jmptbl


	ALIGN(16)
L_mpn_mul_1_main_loop:

L_mpn_mul_1_main_loop_s16:
	mulq	s2limb
	movq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*0)), %rax
	addq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*0-3))
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_mul_1_main_loop_s15:
	mulq	s2limb
	movq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*0+1)), %rax
	addq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*0-2))
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_mul_1_main_loop_s14:
	mulq	s2limb
	movq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*1)), %rax
	addq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*1-3))
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_mul_1_main_loop_s13:
	mulq	s2limb
	movq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*1+1)), %rax
	addq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*1-2))
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_mul_1_main_loop_s12:
	mulq	s2limb
	movq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*2)), %rax
	addq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*2-3))
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_mul_1_main_loop_s11:
	mulq	s2limb
	movq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*2+1)), %rax
	addq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*2-2))
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_mul_1_main_loop_s10:
	mulq	s2limb
	movq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*3)), %rax
	addq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*3-3))
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_mul_1_main_loop_s9:
	mulq	s2limb
	movq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*3+1)), %rax
	addq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*3-2))
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_mul_1_main_loop_s8:
	mulq	s2limb
	movq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*4)), %rax
	addq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*4-3))
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_mul_1_main_loop_s7:
	mulq	s2limb
	movq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*4+1)), %rax
	addq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*4-2))
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_mul_1_main_loop_s6:
	mulq	s2limb
	movq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*5)), %rax
	addq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*5-3))
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_mul_1_main_loop_s5:
	mulq	s2limb
	movq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*5+1)), %rax
	addq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*5-2))
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_mul_1_main_loop_s4:
	mulq	s2limb
	movq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*6)), %rax
	addq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*6-3))
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_mul_1_main_loop_s3:
	mulq	s2limb
	movq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*6+1)), %rax
	addq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*6-2))
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_mul_1_main_loop_s2:
	mulq	s2limb
	movq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*7)), %rax
	addq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*7-3))
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_mul_1_main_loop_s1:
	mulq	s2limb
	movq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*7+1)), %rax
	addq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*7-2))
	leaq	(%rdx), B_y
	adcq	$0, A_y

	addq	$UNROLL_SIZE, index
	jnz	L_mpn_mul_1_main_loop

L_mpn_mul_1_finish:	

	mulq	s2limb
	movq	A_x, ADDR(rp,index,-3)
	addq	A_y, B_x
	adcq	$0, B_y

	movq	B_x, ADDR(rp,index,-2)
	addq	B_y, %rax
	adcq	$0, %rdx

	movq	%rax, ADDR(rp,index,-1)

	movq	%rdx, %rax
L_mpn_mul_1_exit:


	movq	%rax, (rp)

	addq	$BpL, rp
	addq	$BpL, vp
	decq	vn
	jz	L_mpn_mul_basecase_exit



	movq	un, %rdx

	cmp	$4,%rdx
	jge	L_mpn_addmul_1_set_jmptbl

		leaq	L_mpn_addmul_1_jmptbl0(%rip), %rax
		addq	(%rax, %rdx, 8), %rax
		pushq	$0
		pushq	%rax
		jmp	L_mpn_addmul_1_set_jmptbl_end
	L_mpn_addmul_1_set_jmptbl:
		subq	$4,%rdx
		andq	$UNROLL_MASK, %rdx

		leaq	(4-UNROLL_SIZE)(%rdx), %rax
		pushq	%rax

		leaq	L_mpn_addmul_1_jmptbl(%rip), %rax
		addq	(%rax, %rdx, 8), %rax
		pushq	%rax
	L_mpn_addmul_1_set_jmptbl_end:



L_mpn_mul_basecase_loop:


	movq	(vp), s2limb


L_mpn_addmul_1:
C	leaq	ADDR(s1p,un,0), s1p
C	leaq	ADDR(rp,un,0), rp
	xorq	index, index
	subq	un, index

	cmp	$4, un
	jge	L_mpn_addmul_1_main_loop_prep

	jmp	*(%rsp)	
		
C	leaq	L_mpn_addmul_1_jmptbl0(%rip), %rax
C	addq	(%rax, %rdx, 8), %rax
C	jmp	*%rax

	ALIGN(8)
L_mpn_addmul_1_jmptbl0:
	.quad	L_mpn_addmul_1_small_0 - L_mpn_addmul_1_jmptbl0
	.quad	L_mpn_addmul_1_small_1 - L_mpn_addmul_1_jmptbl0
	.quad	L_mpn_addmul_1_small_2 - L_mpn_addmul_1_jmptbl0
	.quad	L_mpn_addmul_1_small_3 - L_mpn_addmul_1_jmptbl0


L_mpn_addmul_1_small_0:
	xorq	%rax, %rax
	jmp	L_mpn_addmul_1_exit

L_mpn_addmul_1_small_1:
	movq	ADDR(s1p,index,0), %rax
	mulq	s2limb
	movq	ADDR(rp,index,0), A_z

	addq	%rax, A_z
	movq	$0, %rax
	movq	A_z, ADDR(rp,index,0)
	adcq	%rdx, %rax

	jmp	L_mpn_addmul_1_exit

L_mpn_addmul_1_small_2:
	movq	ADDR(s1p,index,0), %rax
	mulq	s2limb
	movq	ADDR(rp,index,0), A_z
	movq	%rax, A_x
	movq	%rdx, A_y

	movq	ADDR(s1p,index,1), %rax
	mulq	s2limb
	movq	ADDR(rp,index,1), B_z

	addq	A_x, A_z
	adcq	A_y, %rax
	movq	A_z, ADDR(rp,index,0)
	adcq	$0, %rdx

	addq	%rax, B_z
	movq	$0, %rax
	movq	B_z, ADDR(rp,index,1)
	adcq	%rdx, %rax

	jmp	L_mpn_addmul_1_exit

L_mpn_addmul_1_small_3:
	movq	ADDR(s1p,index,0), %rax
	mulq	s2limb
	movq	ADDR(rp,index,0), A_z
	movq	%rax, A_x
	movq	%rdx, A_y

	movq	ADDR(s1p,index,1), %rax
	mulq	s2limb
	movq	ADDR(rp,index,1), B_z
	movq	%rax, B_x
	movq	%rdx, B_y

	movq	ADDR(s1p,index,2), %rax

	mulq	s2limb
	addq	A_x, A_z
	adcq	A_y, B_x
	movq	A_z, ADDR(rp,index,0)
	movq	ADDR(rp,index,2), A_z
	adcq	$0, B_y

	addq	B_x, B_z
	adcq	B_y, %rax
	movq	B_z, ADDR(rp,index,1)
	adcq	$0, %rdx

	addq	%rax, A_z
	movq	$0, %rax
	movq	A_z, ADDR(rp,index,2)
	adcq	%rdx, %rax

	jmp	L_mpn_addmul_1_exit


L_mpn_addmul_1_main_loop_prep:
C	movq	%rdx, temp

	testq	$1, un
	jz	L_mpn_addmul_1_main_loop_prep_n1

	movq	ADDR(s1p,index,0), %rax
	mulq	s2limb
	movq	ADDR(rp,index,0), A_z
	movq	%rax, A_x
	movq	%rdx, A_y

	movq	ADDR(s1p,index,1), %rax
	mulq	s2limb
	movq	ADDR(rp,index,1), B_z
	movq	%rax, B_x
	movq	%rdx, B_y

	jmp	L_mpn_addmul_1_main_loop_prep_n2

L_mpn_addmul_1_main_loop_prep_n1:
	movq	ADDR(s1p,index,0), %rax
	mulq	s2limb
	movq	ADDR(rp,index,0), B_z
	movq	%rax, B_x
	movq	%rdx, B_y

	movq	ADDR(s1p,index,1), %rax
	mulq	s2limb
	movq	ADDR(rp,index,1), A_z
	movq	%rax, A_x
	movq	%rdx, A_y

L_mpn_addmul_1_main_loop_prep_n2:

	movq	ADDR(s1p,index,2), %rax

	addq	8(%rsp), index
	jmp	*(%rsp)
C	subq	$4, temp
C	andq	$UNROLL_MASK, temp
C	lea	(4-UNROLL_SIZE)(index, temp), index

C	leaq	L_mpn_addmul_1_jmptbl(%rip), %rdx
C	addq	(%rdx, temp, 8), %rdx
C	jmp	*%rdx


	ALIGN(8)
L_mpn_addmul_1_jmptbl:
	.quad	L_mpn_addmul_1_main_loop_s1 - L_mpn_addmul_1_jmptbl
	.quad	L_mpn_addmul_1_main_loop_s2 - L_mpn_addmul_1_jmptbl
	.quad	L_mpn_addmul_1_main_loop_s3 - L_mpn_addmul_1_jmptbl
	.quad	L_mpn_addmul_1_main_loop_s4 - L_mpn_addmul_1_jmptbl
	.quad	L_mpn_addmul_1_main_loop_s5 - L_mpn_addmul_1_jmptbl
	.quad	L_mpn_addmul_1_main_loop_s6 - L_mpn_addmul_1_jmptbl
	.quad	L_mpn_addmul_1_main_loop_s7 - L_mpn_addmul_1_jmptbl
	.quad	L_mpn_addmul_1_main_loop_s8 - L_mpn_addmul_1_jmptbl
	.quad	L_mpn_addmul_1_main_loop_s9 - L_mpn_addmul_1_jmptbl
	.quad	L_mpn_addmul_1_main_loop_s10 - L_mpn_addmul_1_jmptbl
	.quad	L_mpn_addmul_1_main_loop_s11 - L_mpn_addmul_1_jmptbl
	.quad	L_mpn_addmul_1_main_loop_s12 - L_mpn_addmul_1_jmptbl
	.quad	L_mpn_addmul_1_main_loop_s13 - L_mpn_addmul_1_jmptbl
	.quad	L_mpn_addmul_1_main_loop_s14 - L_mpn_addmul_1_jmptbl
	.quad	L_mpn_addmul_1_main_loop_s15 - L_mpn_addmul_1_jmptbl
	.quad	L_mpn_addmul_1_main_loop_s16 - L_mpn_addmul_1_jmptbl


	ALIGN(16)
L_mpn_addmul_1_main_loop:

L_mpn_addmul_1_main_loop_s16:
	mulq	s2limb
	addq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*0)), %rax
	adcq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*0-3))
	movq	ADDR(rp,index,eval(2*0-1)), A_z
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_addmul_1_main_loop_s15:
	mulq	s2limb
	addq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*0+1)), %rax
	adcq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*0-2))
	movq	ADDR(rp,index,eval(2*0)), B_z
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_addmul_1_main_loop_s14:
	mulq	s2limb
	addq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*1)), %rax
	adcq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*1-3))
	movq	ADDR(rp,index,eval(2*1-1)), A_z
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_addmul_1_main_loop_s13:
	mulq	s2limb
	addq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*1+1)), %rax
	adcq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*1-2))
	movq	ADDR(rp,index,eval(2*1)), B_z
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_addmul_1_main_loop_s12:
	mulq	s2limb
	addq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*2)), %rax
	adcq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*2-3))
	movq	ADDR(rp,index,eval(2*2-1)), A_z
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_addmul_1_main_loop_s11:
	mulq	s2limb
	addq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*2+1)), %rax
	adcq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*2-2))
	movq	ADDR(rp,index,eval(2*2)), B_z
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_addmul_1_main_loop_s10:
	mulq	s2limb
	addq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*3)), %rax
	adcq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*3-3))
	movq	ADDR(rp,index,eval(2*3-1)), A_z
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_addmul_1_main_loop_s9:
	mulq	s2limb
	addq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*3+1)), %rax
	adcq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*3-2))
	movq	ADDR(rp,index,eval(2*3)), B_z
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_addmul_1_main_loop_s8:
	mulq	s2limb
	addq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*4)), %rax
	adcq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*4-3))
	movq	ADDR(rp,index,eval(2*4-1)), A_z
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_addmul_1_main_loop_s7:
	mulq	s2limb
	addq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*4+1)), %rax
	adcq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*4-2))
	movq	ADDR(rp,index,eval(2*4)), B_z
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_addmul_1_main_loop_s6:
	mulq	s2limb
	addq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*5)), %rax
	adcq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*5-3))
	movq	ADDR(rp,index,eval(2*5-1)), A_z
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_addmul_1_main_loop_s5:
	mulq	s2limb
	addq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*5+1)), %rax
	adcq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*5-2))
	movq	ADDR(rp,index,eval(2*5)), B_z
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_addmul_1_main_loop_s4:
	mulq	s2limb
	addq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*6)), %rax
	adcq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*6-3))
	movq	ADDR(rp,index,eval(2*6-1)), A_z
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_addmul_1_main_loop_s3:
	mulq	s2limb
	addq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*6+1)), %rax
	adcq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*6-2))
	movq	ADDR(rp,index,eval(2*6)), B_z
	leaq	(%rdx), B_y
	adcq	$0, A_y

L_mpn_addmul_1_main_loop_s2:
	mulq	s2limb
	addq	A_x, A_z
	leaq	(%rax), A_x
	movq	ADDR(s1p,index,eval(2*7)), %rax
	adcq	A_y, B_x
	movq	A_z, ADDR(rp,index,eval(2*7-3))
	movq	ADDR(rp,index,eval(2*7-1)), A_z
	leaq	(%rdx), A_y
	adcq	$0, B_y

L_mpn_addmul_1_main_loop_s1:
	mulq	s2limb
	addq	B_x, B_z
	leaq	(%rax), B_x
	movq	ADDR(s1p,index,eval(2*7+1)), %rax
	adcq	B_y, A_x
	movq	B_z, ADDR(rp,index,eval(2*7-2))
	movq	ADDR(rp,index,eval(2*7)), B_z
	leaq	(%rdx), B_y
	adcq	$0, A_y

	addq	$UNROLL_SIZE, index
	jnz	L_mpn_addmul_1_main_loop

L_mpn_addmul_1_finish:	

	mulq	s2limb
	addq	A_x, A_z
	movq	A_z, ADDR(rp,index,-3)
	movq	ADDR(rp,index,-1), A_z
	adcq	A_y, B_x
	adcq	$0, B_y

	addq	B_x, B_z
	movq	B_z, ADDR(rp,index,-2)
	adcq	B_y, %rax
	adcq	$0, %rdx

	addq	%rax, A_z
	movq	$0, %rax
	movq	A_z, ADDR(rp,index,-1)
	adcq	%rdx, %rax

L_mpn_addmul_1_exit:

	movq	%rax, (rp)

	addq	$BpL, rp
	addq	$BpL, vp
	decq	vn
	jnz	L_mpn_mul_basecase_loop

	addq	$(8*2), %rsp



L_mpn_mul_basecase_exit:
	popq	%rbp
	popq	%rbx
	popq	%r12
	popq	%r13
	popq	%r14
	popq	%r15
	ret
EPILOGUE()
